#!/bin/bash


script_dir=$(dirname $0)
bin_dir=$script_dir/../bin

encoder=$bin_dir/col2csp.py
write_proj=$bin_dir/write_projection.py
mc_solver=gpmc
mc_mode="-mode=2"

ans_regex="^c s exact arb int ([0-9]+)"

max_model_num=3

usage() {
    echo "
usage:
    $0 [-h] [-v] [-m <int>[1..$max_model_num]] [-d <str>] <COLFILE>
" >&2
    exit 1
}

flag_v=false
model_num=2
flag_d=false
directory=""

while getopts vm:d:h OPT; do
    case $OPT in
    v)
        flag_v=true
        ;;
    m)
	model_num=$OPTARG
	;;
    d)
	flag_d=true
	directory=$OPTARG
	;;
    h | \?)
        usage && exit 1
        ;;
    esac
done

shift $((OPTIND - 1))

if [[ $# -ne 1 ]]; then
    usage
fi

if $flag_d; then
    mkdir -p $directory/csp
    mkdir -p $directory/cnf
    mkdir -p $directory/map
    mkdir -p $directory/ecnf
    mkdir -p $directory/log
    file_name=${1##*/}
    file_name=${file_name%%.*}
    csp_file=$directory/csp/$file_name\_$$.csp
    sat_file=$directory/cnf/$file_name\_$$.cnf
    map_file=$directory/map/$file_name\_$$.map
    ecnf_file=$directory/ecnf/$file_name\_$$.ecnf
    log_file=$directory/log/$file_name\_$$.log
else
    tmp_file=$(mktemp)
    csp_file=$tmp_file.csp
    sat_file=$tmp_file.cnf
    map_file=$tmp_file.map
    ecnf_file=$tmp_file.ecnf
    log_file=$tmp_file.log
fi

function show_cmd {
    if $flag_v; then
        echo "c"
        echo "c ------------------------------------------------------------"
        echo "c CMD $1"
        echo "c ------------------------------------------------------------"
    fi
}

function verbose {
    if $flag_v; then
        sed "s/^/c /g" $1
    fi
}

function check_colfile {
    if [ ! -f $1 -o ! -s $1 ]; then
        echo "COLFILE ERROR: $1 is not exsist or is empty file."
        usage
    fi
}

function output_answer {
    while read line
    do
	if [[ $line =~ $ans_regex ]]; then
	    echo ${BASH_REMATCH[1]}
	    break
	fi
    done < $1
}

function rm_tmpfile {
  [[ -f "$1" ]] && rm -f "$1"
}

function finally {
    if ! $flag_d; then
	rm_tmpfile $tmp_file
	rm_tmpfile $csp_file
	rm_tmpfile $sat_file
	rm_tmpfile $map_file
	rm_tmpfile $ecnf_file
	rm_tmpfile $log_file
    fi
}

trap 'finally' EXIT
trap 'finally; exit 1' INT PIPE TERM

check_colfile $1

show_cmd "python3 $encoder -m $model_num $1 > $csp_file"
python3 $encoder -m $model_num $1 > $csp_file
verbose $csp_file

show_cmd "sugar -jopt1=\"-Xmx16g\" $csp_file -sat $sat_file -map $map_file -n"
sugar $csp_file -jopt1="-Xmx16g" -sat $sat_file -map $map_file -n
verbose $map_file

show_cmd "python3 $write_proj $sat_file $map_file > $ecnf_file"
python3 $write_proj $sat_file $map_file > $ecnf_file

show_cmd "$mc_solver $mc_mode $ecnf_file > $log_file"
$mc_solver $mc_mode $ecnf_file > $log_file

verbose $log_file
output_answer $log_file
