from util_csp import *
import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import shortest_path


"""
Prepro
"""

def _deg(i, edges):
    return len([0 for e in edges if e[0] == i or e[1] == i])

def _prepro_terminal_degree(edges, start, goal):
    if _deg(start, edges) < _deg(goal, edges):
        pass
    else:
        tmp = start
        start = goal
        goal = tmp
    print_comment("Start:", start)
    print_comment("Goal:", goal)
    return start, goal

dist = []

def _dist(i, j):
    return int(dist[i-1][j-1])

def _prepro_dist(nodes, edges):
    n = len(nodes)
    edge = np.array([[e[0], e[1], 1] for e in edges], dtype=np.int64).T
    graph = csr_matrix((edge[2], (edge[:2] - 1)), (n, n))
    global dist
    dist = shortest_path(graph, method='auto', directed=False)

deadend = []

def _prepro_deadend(nodes, edges, start, goal):
    new_nodes = nodes.copy()
    new_edges = edges.copy()
    deadend_nodes = []
    check_nodes = nodes.copy()
    while check_nodes:
        tmp_nodes = [i for i in new_nodes if i not in check_nodes]
        tmp_edges = []
        next_check_nodes = []
        for i in check_nodes:
            if i not in [start, goal] and _deg(i,new_edges) <= 1:
                deadend_nodes.append(i)
            else:
                tmp_nodes.append(i)
        for e in new_edges:
            if e[0] in tmp_nodes and e[1] in tmp_nodes:
                tmp_edges.append(e)
            elif e[0] not in tmp_nodes:
                if e[1] not in next_check_nodes:
                    next_check_nodes.append(e[1])
            elif e[1] not in tmp_nodes:
                if e[0] not in next_check_nodes:
                    next_check_nodes.append(e[0])
        new_nodes = tmp_nodes.copy()
        new_edges = tmp_edges.copy()
        check_nodes = next_check_nodes.copy()
    global deadend
    deadend = deadend_nodes.copy()

def _prepro_node_reduction(nodes, edges, start, goal, max_length):
    new_nodes = []
    new_edges = []
    removed_nodes = []
    removed_edges = []
    for i in nodes:
        if _dist(start,i) + _dist(i,goal) <= max_length and i not in deadend:
            new_nodes.append(i)
        else:
            removed_nodes.append(i)
    for e in edges:
        i = e[0]
        j = e[1] 
        if max_length >= 1 + min([_dist(start,i)+_dist(j,goal), _dist(start,j)+_dist(i,goal)]) and i in new_nodes and j in new_nodes:
            new_edges.append(e)
        else:
            removed_edges.append(e)
    print_comment(" Removed nodes:", ", ".join([str(i) for i in sorted(removed_nodes)]))
    print_comment(" Removed edges:", ", ".join([f"{e[0]}-{e[1]}" for e in sorted(removed_edges)]))
    print_comment(" Total number of removed nodes:", len(removed_nodes))
    print_comment(" Total number of removed edges:", len(removed_edges))
    return new_nodes, new_edges

def _prepro_edge_direction(edges):
    directed_edges = []
    for e in edges:
        directed_edges += [e,(e[1],e[0])]
    return directed_edges


"""
One pair: Basic modeling
"""

def initialization_one_pair_basic(edges, start, goal):
    start, goal = _prepro_terminal_degree(edges, start, goal)
    edges = _prepro_edge_direction(edges)
    return edges, start, goal
    
def edge_variables_one_pair_basic(edges):
    print_comment("Edge variables")
    for e in edges:
        integer_variable(f"x_{e[0]}_{e[1]}", 0, 1)

def node_label_variables_one_pair_basic(nodes, start, goal, max_length):
    print_comment("Node-label variables")
    for i in nodes:
        if i == start:
            integer_variable(f"y_{i}", 0)
        elif i == goal:
            integer_variable(f"y_{i}", max_length)
        else:
            integer_variable(f"y_{i}", 1, max_length-1)

def indegree_constraints_of_the_start_node_one_pair_basic(edges, start):
    print_comment("In-degree constraints of the start node")
    xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[1] == start]
    print(Eq(Add(*xs), 0))

def outdegree_constraints_of_the_start_node_one_pair_basic(edges, start):
    print_comment("Out-degree constraints of the start node")
    xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[0] == start]
    print(Eq(Add(*xs), 1))

def indegree_constraints_of_the_goal_node_one_pair_basic(edges, goal):
    print_comment("In-degree constraints of the goal node")
    xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[1] == goal]
    print(Eq(Add(*xs), 1)) 

def outdegree_constraints_of_the_goal_node_one_pair_basic(edges, goal):
    print_comment("Out-degree constraints of the goal node")
    xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[0] == goal]
    print(Eq(Add(*xs), 0)) 

def indegree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal):
    print_comment("In-degree constraints of non-terminal nodes")
    for i in nodes:
        if i in [start, goal]:
            continue
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[1] == i]
        print(Le(Add(*xs), 1))

def outdegree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal):
    print_comment("Out-degree constraints of non-terminal nodes")
    for i in nodes:
        if i in [start, goal]:
            continue
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[0] == i]
        print(Le(Add(*xs), 1))

def degree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal):
    print_comment("Degree constraints of non-terminal nodes")
    for i in nodes:
        if i in [start, goal]:
            continue
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[1] == i]
        ys = [f"x_{e[0]}_{e[1]}" for e in edges if e[0] == i]
        print(Eq(Add(*xs), Add(*ys)))

def connectiviy_constraints_one_pair_basic(edges):
    print_comment("Connectivity constraints")
    for e in edges:
        print(Imp(Eq(f"x_{e[0]}_{e[1]}", 1), Lt(f"y_{e[0]}", f"y_{e[1]}")))

def path_length_constraints_one_pair_basic(edges, max_length):
    print_comment("Path length constraints")
    allpath = [f"x_{e[0]}_{e[1]}" for e in edges]
    Ge_bb(max_length, *allpath)

        
"""
One pair: Improved modeling
"""

def initialization_one_pair_improved(nodes, edges, start, goal, max_length):
    start, goal = _prepro_terminal_degree(edges, start, goal)
    _prepro_dist(nodes, edges)
    _prepro_deadend(nodes, edges, start, goal)
    nodes, edges = _prepro_node_reduction(nodes, edges, start, goal, max_length)
    edges = _prepro_edge_direction(edges)
    return nodes, edges, start, goal

def node_label_variables_one_pair_improved(nodes, start, goal, max_length):
    print_comment("Node-label variables")
    for i in nodes:
        if i == start:
            integer_variable(f"y_{i}", 0)
        elif i == goal:
            integer_variable(f"y_{i}", max_length)
        else:
            integer_variable(f"y_{i}", _dist(start,i), max_length-_dist(i,goal))

def node_variables_one_pair_improved(nodes):
    print_comment("Node variables")
    for i in nodes:
        integer_variable(f"z_{i}", 0, 1)

def channeling_constraints_between_node_variables_and_edge_variables_one_pair_improved(edges):
    print_comment("Channeling constraints between node variables and edge variables")
    for e in edges:
        print(Imp(Eq(f"x_{e[0]}_{e[1]}", 1), Eq(f"z_{e[0]}", 1)))
        print(Imp(Eq(f"x_{e[0]}_{e[1]}", 1), Eq(f"z_{e[1]}", 1)))

def channeling_constraints_between_node_variables_and_node_label_variables_one_pair_improved(nodes, start, goal):
    print_comment("Channeling constraints between node variables and node-label variables")
    for i in nodes:
        if i in [start, goal]:
            continue
        print(Imp(Eq(f"z_{i}", 0), Eq(f"y_{i}", _dist(start,i))))
        
def in_path_node_constraints_one_pair_improved(nodes, max_length):
    print_comment("In-path node constraints")
    allnodes = [f"z_{i}" for i in nodes]
    Ge_bb(max_length+1, *allnodes)

def forbidden_node_pair_constraints_one_pair_improved(nodes, start, goal, max_length):
    print_comment("Forbidden node-pair constraints")
    for i in nodes:
        for j in nodes:
            if i < j and max_length < _dist(i,j) + min([_dist(start,i)+_dist(j,goal), _dist(start,j)+_dist(i,goal)]):
                    print(Not(And(Eq(f"z_{i}", 1), Eq(f"z_{j}", 1))))


"""
All pairs: Basic modeling
"""

def initialization_all_pairs_basic(edges):
    edges = _prepro_edge_direction(edges)
    return edges

def start_node_variables_all_pairs_basic(nodes):
    print_comment("Start node variables")
    integer_variable("s", 1, len(nodes))

def goal_node_variables_all_pairs_basic(nodes):
    print_comment("Goal node variables")
    integer_variable("t", 1, len(nodes))

def terminal_constraints_all_pairs_basic():
    print_comment("Terminal constraints")
    print(Lt("s", "t"))

def edge_variables_all_pairs_basic(edges):
    print_comment("Edge variables")
    for e in edges:
        integer_variable(f"x_{e[0]}_{e[1]}", 0, 1)

def node_label_variables_all_pairs_basic(nodes, max_length):
    print_comment("Node-label variables")
    for i in nodes:
        integer_variable(f"y_{i}", 0, max_length)

def indegree_constraints_all_pairs_basic(nodes, edges):
    print_comment("In-degree constraints")
    for i in nodes:
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[1] == i]
        print(Le(Add(*xs), 1))

def outdegree_constraints_all_pairs_basic(nodes, edges):
    print_comment("Out-degree constraints")
    for i in nodes:
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[0] == i]
        print(Le(Add(*xs), 1))
     
def indegree_constraints_of_the_start_node_all_pairs_basic(nodes, edges):
    print_comment("In-degree constraints of the start node")
    for i in nodes:
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[1] == i]
        print(Imp(Eq("s", f"{i}"), Le(Add(*xs), 0)))

def outdegree_constraints_of_the_start_node_all_pairs_basic(nodes, edges):
    print_comment("Out-degree constraints of the start node")
    for i in nodes:
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[0] == i]
        print(Imp(Eq("s", f"{i}"), Ge(Add(*xs), 1)))
        
def indegree_constraints_of_the_goal_node_all_pairs_basic(nodes, edges):
    print_comment("In-degree constraints of the goal node")
    for i in nodes:
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[1] == i]
        print(Imp(Eq("t", f"{i}"), Ge(Add(*xs), 1)))

def outdegree_constraints_of_the_goal_node_all_pairs_basic(nodes, edges):
    print_comment("Out-degree constraints of the goal node")
    for i in nodes:
        xs = [f"x_{e[0]}_{e[1]}" for e in edges if e[0] == i]
        print(Imp(Eq("t", f"{i}"), Le(Add(*xs), 0)))

def degree_constraints_of_nonterminal_nodes_all_pairs_basic(nodes, edges):
    print_comment("Degree constraints of non-terminal nodes")
    for i in nodes:
        xs = [Neq("s", f"{i}"), Neq("t", f"{i}")]
        ys = [f"x_{e[0]}_{e[1]}" for e in edges if e[1] == i]
        zs = [f"x_{e[0]}_{e[1]}" for e in edges if e[0] == i]
        print(Imp(And(*xs), Eq(Add(*ys), Add(*zs))))

def connectiviy_constraints_all_pairs_basic(edges):
    print_comment("Connectivity constraints")
    for e in edges:
        print(Imp(Eq(f"x_{e[0]}_{e[1]}", 1), Lt(f"y_{e[0]}", f"y_{e[1]}")))

def path_length_constraints_all_pairs_basic(edges, max_length):
    print_comment("Path length constraints")
    allpath = [f"x_{e[0]}_{e[1]}" for e in edges]
    Ge_bb(max_length, *allpath)


"""
All pairs: improved modeling
"""

def initialization_all_pairs_improved(nodes, edges):
    _prepro_dist(nodes, edges)
    edges = _prepro_edge_direction(edges)
    return edges

def terminal_combination_constraints_all_pairs_improved(nodes, edges, max_length):
    print_comment("Terminal combination constraints")
    for i in nodes:
        for j in nodes:
            if i < j and max_length < _dist(i,j):
                print(Not(And(Eq("s", i), Eq("t", j))))

def node_variables_all_pairs_improved(nodes):
    print_comment("Node variables")
    for i in nodes:
        integer_variable(f"z_{i}", 0, 1)

def channeling_constraints_between_node_variables_and_edge_variables_all_pairs_improved(edges):
    print_comment("Channeling constraints between node variables and edge variables")
    for e in edges:
        print(Imp(Eq(f"x_{e[0]}_{e[1]}", 1), Eq(f"z_{e[0]}", 1)))
        print(Imp(Eq(f"x_{e[0]}_{e[1]}", 1), Eq(f"z_{e[1]}", 1)))

def channeling_constraints_between_node_variables_and_node_label_variables_all_pairs_improved(nodes):
    print_comment("Channeling constraints between node variables and node-label variables")
    for i in nodes:
        print(Imp(Eq(f"z_{i}", 0), Eq(f"y_{i}", 0)))

def in_path_node_constraints_all_pairs_improved(nodes, max_length):
    print_comment("In-path node constraints")
    allnodes = [f"z_{i}" for i in nodes]
    Ge_bb(max_length+1, *allnodes)

def forbidden_node_pair_constraints_all_pairs_improved(nodes, max_length):
    print_comment("Forbidden node-pair constraints")
    for i in nodes:
        for j in nodes:
            if i < j and max_length < _dist(i,j):
                print(Not(And(Eq(f"z_{i}", 1), Eq(f"z_{j}", 1))))

# only model 3
def forbidden_edge_constraints_all_pairs_improved(nodes, edges, max_length):
    print_comment("Forbidden node-pair constraints")
    for i in nodes:
        for j in nodes:
            if i < j and max_length >= _dist(i,j):
                for e in edges:
                    if max_length < 1 + _dist(i,e[0]) + _dist(e[1],j):
                        print(Not(And(Eq(f"s", i), Eq(f"t", j), Eq(f"x_{e[0]}_{e[1]}", 1))))

"""
print
"""

def print_message(*args):
    elems = [str(x) for x in args]
    print(" ".join(elems), flush=True)

def print_comment(*args):
    print_message(';', *args)
