#!/usr/bin/env python3

"""
@author Shuji Kosuge
"""

import sys
import getopt
import re


class Config:
    def __init__(self, verbose=False):
        self.verbose = verbose


config = Config()


def usage(msg=""):
    if msg != "":
        print(f"{msg}\n")
    print(f"Usage: python3 {sys.argv[0]} <options> cnf_file map_file")
    print(f"    -h --help               : show this help")
    print(f"    -v --verbose            : show detail")
    exit()


def main():
    global config
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hv", ["help", "verbose"])

        for opt, arg in opts:
            if opt in ("-h", "--help"):
                usage()
            elif opt in ("-v", "--verbose"):
                config.verbose = True

        if len(args) != 2:
            usage("please input cnf_file and map_file.")

    except getopt.GetoptError as err:
        usage(str(err))

    cnf_file = args[0]
    map_file = args[1]

    projection = []
    with open(map_file) as mf:
        for line in mf:
            m = re.match("^int\s(x_\d+_\d+)\s(\d+)\s0\.\.1", line)
            if m is not None:
                projection.append(m.group(2))
    with open(cnf_file) as cf:
        for line in cf:
            print(line, end="")
            p = re.match("^p\scnf\s\d+\s\d+", line)
            if p is not None:
                print(f"c p show {' '.join(projection)} 0")


if __name__ == "__main__":
    main()
